@@ -42,10 +42,15 @@ gettext.bindtextdomain(APP, LOCALE_DIR)
 gettext.textdomain(APP)
 _ = gettext.gettext
 
-class Application():
+class Application:
 
     def __init__(self):
 
+        self.test_mode = False
+        if len(sys.argv) > 1 and sys.argv[1] == "test":
+            self.test_mode = True
+            print("Test mode detected, adding a dummy device.")
+
         self.builder = Gtk.Builder()
         self.builder.set_translation_domain(APP)
         self.builder.add_from_file("/usr/share/linuxmint/mintdrivers/main.ui")
@@ -63,9 +68,6 @@ class Application():
 
         self.window_main.connect("delete_event", self.quit_application)
 
-        self.stack = self.builder.get_object("stack")
-        self.spinner = self.builder.get_object("spinner")
-
         self.button_driver_revert = Gtk.Button(label=_("Re_vert"), use_underline=True)
         self.button_driver_revert.connect("clicked", self.on_driver_changes_revert)
         self.button_driver_apply = Gtk.Button(label=_("_Apply Changes"), use_underline=True)
@@ -86,8 +88,8 @@ class Application():
         self.box_driver_action.pack_end(self.button_driver_cancel, False, False, 0)
 
         self.builder.get_object("error_button").connect("clicked", self.on_error_button)
-
-        self.info_bar.set_no_show_all(True)
+        self.builder.get_object("button_mount_media").connect("clicked", self.on_mount_media_button)
+        self.builder.get_object("button_offline").connect("clicked", self.check_internet_or_live_media)
 
         self.progress_bar = Gtk.ProgressBar(valign=Gtk.Align.CENTER)
         self.box_driver_action.pack_end(self.progress_bar, True, True, 0)
@@ -96,54 +98,66 @@ class Application():
         self.needs_restart = False
         self.live_mode = False
 
+        self.show_page("refresh_page")
+
         with open('/proc/cmdline') as f:
             cmdline = f.read()
-            if ((not "boot=casper" in cmdline) and (not "boot=live" in cmdline)):
-                print ("Post-install mode detected")
-                self.info_bar_label.set_text(_("Drivers cannot be installed.\nPlease connect to the Internet or insert the Linux Mint installation DVD (or USB stick)."))
-                self.check_internet_or_live_dvd()
-                self.button_info_bar.connect("clicked", self.check_internet_or_live_dvd)
-            else:
+            if ("boot=casper" in cmdline) or ("boot=live" in cmdline):
                 print ("Live mode detected")
                 self.live_mode = True
-                self.info_bar.hide()
                 self.update_cache()
+            else:
+                self.check_internet_or_live_media()
 
+    def show_page(self, page):
+        if page == "refresh_page":
+            self.builder.get_object("spinner").start()
+        else:
+            self.builder.get_object("spinner").stop()
+        self.builder.get_object("stack").set_visible_child_name(page)
 
     def on_error_button(self, button):
-        self.stack.set_visible_child_name("drivers_page")
-        self.spinner.stop()
+        self.show_page("drivers_page")
 
     def update_cache(self):
-        self.stack.set_visible_child_name("refresh_page")
-        self.spinner.start()
+        print("Updating cache")
+        self.show_page("refresh_page")
         task = packagekit.Task()
         task.refresh_cache_async(True, Gio.Cancellable(), self.on_cache_update_progress, (None, ), self.on_cache_update_finished, (None, ))
 
-    def on_error(self, msg):
-        self.stack.set_visible_child_name("error_page")
-        self.spinner.stop()
-        self.builder.get_object("error_label").set_label(msg)
+    def on_error(self, error):
+        # Returns False if the error was from cancelling or failing to authenticate.
+        # This will bring the ui back to pre-apply state. Returning True will reset
+        # entirely.
+
+        # it thinks it's a PkClientError but it's really PkErrorEnum
+        # the GError code is set to 0xFF + code
+        real_code = error.code
+        if error.code >= 0xFF:
+            real_code = error.code - 0xFF
+
+            if real_code == packagekit.ErrorEnum.NOT_AUTHORIZED:
+                # Silently ignore auth failures or cancellation.
+                return False
+
+        self.show_page("error_page")
+        self.builder.get_object("error_label").set_label(error.message)
+        return True
 
     def on_cache_update_progress(self, progress, ptype, data=None):
         pass
 
     def on_cache_update_finished(self, source, result, data=None):
+        print("Cache updated")
         XApp.set_window_progress(self.window_main, 0)
         self.get_drivers_async()
 
     def quit_application(self, widget=None, event=None):
-        self.clean_up_media_cdrom()
+        self.cleanup_live_media()
         Gtk.main_quit()
 
-    def clean_up_media_cdrom(self):
-        if os.path.exists("/media/cdrom"):
-            print ("unmounting /media/cdrom...")
-            os.system("umount /media/cdrom")
-            print ("removing /media/cdrom...")
-            os.system("rm -rf /media/cdrom")
-            print ("removing cdrom repository...")
-            os.system("sed -i '/deb cdrom/d' /etc/apt/sources.list")
+    def cleanup_live_media(self):
+        subprocess.call(["sudo", "mintdrivers-remove-live-media"])
 
     def check_connectivity(self, reference):
         try:
@@ -152,76 +166,55 @@ class Application():
         except:
             return False
 
-    def check_internet_or_live_dvd(self, widget=None):
-
+    @idle
+    def check_internet_or_live_media(self, widget=None):
+        self.show_page("refresh_page")
+        print ("Checking Internet connectivity...")
         try:
-            print ("Checking Internet connectivity...")
+            urllib.request.urlopen("http://archive.ubuntu.com", timeout=10)
+            # We're online
+            print ("  --> Computer is online")
+            self.update_cache()
+            return
+        except:
+            print ("  --> Computer is offline")
 
-            # We either get drivers from the Internet of from the liveDVD
-            # So check the connection to the Internet or scan for a liveDVD
+        # We're offline, let's look for a live media
+        print ("Checking for a live media...")
+        mount_point = None
+        partitions = psutil.disk_partitions()
+        for p in partitions:
+            if p.fstype == "iso9660":
+                mount_point = p.mountpoint
+                print ("  --> Found: %s at %s" % (p.device, p.mountpoint))
+                break
 
-            if self.check_connectivity("http://archive.ubuntu.com"):
-            # We can reach the repository, everything's fine
-                print ("  --> Internet connection detected.")
-                self.info_bar.hide()
-                self.update_cache()
-                return
-            else:
-                print ("  --> Internet connection is missing.")
+        if mount_point is None:
+            # Offline and no live media, show the offline page
+            print ("  --> None found.")
+            self.show_page("offline_page")
+            return
 
+        # We're offline but an ISO was detected
+        # Let's make sure it's mounted as a repository
+        if os.path.exists("/media/mintdrivers/README.diskdefines"):
+            print ("  --> Mounted in /media/mintdrivers")
+            self.update_cache()
+        else:
+            print ("  --> Not mounted in /media/mintdrivers")
+            self.show_page("media_page")
 
-            print ("Checking access to live media...")
+    def on_mount_media_button(self, button):
+        print("Mounting live media")
+        self.show_page("refresh_page")
+        self.mount_live_media()
 
-            # We can't reach the repository, we need the installation media
-            mount_point = None
-            mounted_on_media_cdrom = False
-
-            # Find the live media
-            partitions = psutil.disk_partitions()
-            for p in partitions:
-                try:
-                    if p.fstype == "iso9660":
-                        mount_point = p.mountpoint
-                        print ("Found live media: %s at %s" % (p.device, p.mountpoint))
-                        # Add it to apt-cdrom
-                        p = subprocess.Popen(["sudo", "apt-cdrom", "-d", mount_point, "-m", "add"], stderr=subprocess.PIPE)
-                        (stdout, stderr) = p.communicate()
-                        mounted_on_media_cdrom = True
-                        if 'E: ' in stderr.decode():
-                           mounted_on_media_cdrom = False
-                        break
-                except Exception as e:
-                    print ("Exception while calling apt-cdrom:", e)
-
-            if mount_point is None:
-                # Not mounted..
-                print ("No live media found.")
-                self.info_bar.show()
-                self.stack.set_visible_child_name("drivers_page")
-                self.spinner.stop()
-                return
-
-            if mounted_on_media_cdrom == False or not os.path.exists("/media/cdrom/README.diskdefines"):
-                # Mounted, but not in the right place...
-                print ("Binding mount %s to /media/cdrom" % mount_point)
-                self.clean_up_media_cdrom()
-                os.system("mkdir -p /media/cdrom")
-                p = subprocess.Popen(["mount", "--bind", mount_point, "/media/cdrom"], stderr=subprocess.PIPE)
-                stderr = p.communicate()
-
-            # It should now be mounted to /media/cdrom, else something went wrong..
-            if os.path.exists("/media/cdrom/README.diskdefines"):
-                print ("Live media detected")
-                self.info_bar.hide()
-                self.update_cache()
-            else:
-                self.info_bar.show()
-        except Exception as e:
-            print("An error occurred: {}".format(e))
-            self.info_bar.show()
+    @_async
+    def mount_live_media(self):
+        subprocess.call(["/usr/bin/pkexec", "mintdrivers-add-live-media"])
+        self.check_internet_or_live_media()
 
     def on_driver_changes_progress(self, progress, ptype, data=None):
-        #print(progress)
         self.button_driver_revert.set_visible(False)
         self.button_driver_apply.set_visible(False)
         self.button_driver_restart.set_visible(False)
@@ -229,25 +222,41 @@ class Application():
         self.progress_bar.set_visible(True)
         self.progress_bar.set_visible(True)
 
-        self.label_driver_action.set_label(_("Applying changes..."))
+        if progress.get_status() == packagekit.StatusEnum.DOWNLOAD:
+            self.label_driver_action.set_label(_("Downloading drivers..."))
+        elif progress.get_status() == packagekit.StatusEnum.INSTALL:
+            self.label_driver_action.set_label(_("Installing drivers..."))
+        elif progress.get_status() == packagekit.StatusEnum.REMOVE:
+            self.label_driver_action.set_label(_("Removing drivers..."))
+        elif progress.get_status() == packagekit.StatusEnum.CANCEL:
+            self.label_driver_action.set_label(_("Cancelling..."))
+        elif progress.get_status() == packagekit.StatusEnum.LOADING_CACHE:
+            self.label_driver_action.set_label(_("Loading cache..."))
+        else:
+            self.label_driver_action.set_label("")
         if ptype == packagekit.ProgressType.PERCENTAGE:
             prog_value = progress.get_property('percentage')
             self.progress_bar.set_fraction(prog_value / 100.0)
             XApp.set_window_progress(self.window_main, prog_value)
 
-    def on_driver_changes_finish(self, source, result, installs_pending):
+    def on_driver_changes_finish(self, source, result, installs):
         results = None
         errors = False
         try:
             results = self.pk_task.generic_finish(result)
-        except Exception as e:
-            self.on_driver_changes_revert()
-            self.on_error(str(e))
+        except GLib.Error as e:
             errors = True
-        if not installs_pending:
+            if self.on_error(e):
+                # real failure
+                self.on_driver_changes_revert()
+                self.clear_changes()
+            else:
+                self.button_driver_revert.set_sensitive(bool(self.driver_changes))
+                self.button_driver_apply.set_sensitive(bool(self.driver_changes))
+
+        if installs is None or len(installs) == 0 or errors:
             self.needs_restart = (not errors)
             self.progress_bar.set_visible(False)
-            self.clear_changes()
             self.apt_cache = apt.Cache()
             self.set_driver_action_status()
             self.update_label_and_icons_from_status()
@@ -255,7 +264,16 @@ class Application():
             self.button_driver_apply.set_visible(True)
             self.button_driver_cancel.set_visible(False)
             self.scrolled_window_drivers.set_sensitive(True)
-        XApp.set_window_progress(self.window_main, 0)
+            XApp.set_window_progress(self.window_main, 0)
+        else:
+            print("Installing", installs)
+            self.pk_task.install_packages_async(installs,
+                    self.cancellable,  # cancellable
+                    self.on_driver_changes_progress,
+                    (None, ),  # progress data
+                    self.on_driver_changes_finish,  # GAsyncReadyCallback
+                    None  # callback data
+             )
 
     def on_driver_changes_apply(self, button):
         self.pk_task = packagekit.Task()
@@ -278,26 +296,39 @@ class Application():
 
         self.cancellable = Gio.Cancellable()
         try:
-            if removals:
-                installs_pending = False
-                if installs:
-                    installs_pending = True
-                self.pk_task.remove_packages_async(removals,
-                            False,  # allow deps
-                            True,  # autoremove
-                            self.cancellable,  # cancellable
-                            self.on_driver_changes_progress,
-                            (None, ),  # progress data
-                            self.on_driver_changes_finish,  # callback ready
-                            installs_pending  # callback data
-                 )
-            if installs:
+            if len(removals) > 0:
+                try:
+                    # Try to purge (Mint specific version of packagekit)
+                    print("Purging", removals)
+                    self.pk_task.purge_packages_async(removals,
+                                False,  # allow deps
+                                True,  # autoremove
+                                self.cancellable,  # cancellable
+                                self.on_driver_changes_progress,
+                                (None, ),  # progress data
+                                self.on_driver_changes_finish,  # callback ready
+                                installs  # callback data
+                     )
+                except:
+                    # If purging isn't supported, just remove
+                    print("Couldn't purge! Removing", removals)
+                    self.pk_task.remove_packages_async(removals,
+                                False,  # allow deps
+                                True,  # autoremove
+                                self.cancellable,  # cancellable
+                                self.on_driver_changes_progress,
+                                (None, ),  # progress data
+                                self.on_driver_changes_finish,  # callback ready
+                                installs  # callback data
+                     )
+            elif len(installs) > 0:
+                print("Installing", installs)
                 self.pk_task.install_packages_async(installs,
                         self.cancellable,  # cancellable
                         self.on_driver_changes_progress,
                         (None, ),  # progress data
                         self.on_driver_changes_finish,  # GAsyncReadyCallback
-                        False  # ready data
+                        None  # callback data
                  )
 
             self.button_driver_revert.set_sensitive(False)
@@ -327,7 +358,7 @@ class Application():
         self.clear_changes()
 
     def on_driver_restart_clicked(self, button_restart):
-        self.clean_up_media_cdrom()
+        self.cleanup_live_media()
         subprocess.call(['systemctl', 'reboot'])
 
     def clear_changes(self):
@@ -388,7 +419,7 @@ class Application():
         return dependencies
 
     def gather_device_data(self, device):
-        '''Get various device data used to build the GUI.
+        """Get various device data used to build the GUI.
 
           return a tuple of (overall_status string, icon, drivers dict).
           the drivers dict is using this form:
@@ -404,7 +435,7 @@ class Application():
 
              Please note that either manually_installed and no_driver are set to None if not applicable
              (no_driver isn't present if there are builtins)
-        '''
+        """
 
         possible_overall_status = {
             'recommended': (_("This device is using the recommended driver."), "recommended-driver"),
@@ -423,13 +454,26 @@ class Application():
         except KeyError:
             pass
 
+        # -open nvidia drivers are recommended now over normal ones. Go thru the list and get the version of the recommended one,
+        # then we can flag the non-'open' one instead.
+        new_recommended = None
+        for pkg_driver_name in device['drivers']:
+            current_driver = device['drivers'][pkg_driver_name]
+            try:
+                if current_driver['recommended'] and current_driver['from_distro']:
+                    driver_status = 'recommended'
+                    if pkg_driver_name.endswith("-open"):
+                        new_recommended = pkg_driver_name.replace("-open", "")
+            except KeyError:
+                pass
+
         for pkg_driver_name in device['drivers']:
             current_driver = device['drivers'][pkg_driver_name]
 
             # get general status
             driver_status = 'alternative'
             try:
-                if current_driver['recommended'] and current_driver['from_distro']:
+                if (current_driver['recommended'] and current_driver['from_distro']) or pkg_driver_name == new_recommended:
                     driver_status = 'recommended'
             except KeyError:
                 pass
@@ -445,9 +489,15 @@ class Application():
             try:
                 pkg = self.apt_cache[pkg_driver_name]
                 installed = pkg.is_installed
+                if installed:
+                    version = pkg.installed.version
+                    summary = pkg.installed.summary
+                else:
+                    version = pkg.candidate.version
+                    summary = pkg.candidate.summary
                 description_line1 = "<b>%s</b>" % pkg.shortname
-                description_line2 = "<small>%s</small> %s" % (_("Version"), pkg.candidate.version)
-                description_line3 = "<small>%s</small>" % pkg.candidate.summary
+                description_line2 = "<small>%s</small> %s" % (_("Version"), version)
+                description_line3 = "<small>%s</small>" % summary
                 if driver_status == 'recommended':
                     description_line1 = "%s <b><small><span foreground='#58822B'>(%s)</span></small></b>" % (description_line1, _("recommended"))
                 if current_driver['free'] and pkg.shortname != "bcmwl-kernel-source" and (not pkg.shortname.startswith("nvidia-")):
@@ -493,7 +543,7 @@ class Application():
                 if returned_drivers[section][keys]['selected']:
                     (overall_status, icon) = possible_overall_status[section]
 
-        return (overall_status, icon, returned_drivers)
+        return overall_status, icon, returned_drivers
 
     def get_device_icon(self, device):
         vendor = device.get('vendor', _('Unknown'))
@@ -513,7 +563,7 @@ class Application():
         elif "amd64-microcode" in device['drivers']:
             icon = "amd"
 
-        return (GdkPixbuf.Pixbuf.new_from_file_at_size("/usr/share/linuxmint/mintdrivers/icons/%s.svg" % icon, 48, -1))
+        return GdkPixbuf.Pixbuf.new_from_file_at_size("/usr/share/linuxmint/mintdrivers/icons/%s.svg" % icon, 48, -1)
 
     def get_cpu_name(self):
         with open("/proc/cpuinfo") as cpuinfo:
@@ -526,6 +576,16 @@ class Application():
     def get_drivers_async(self):
         self.apt_cache = apt.Cache()
         self.devices = detect.system_device_drivers()
+        if self.test_mode:
+            dummy_device = {
+                'modalias': '',
+                'vendor': 'Linux Mint', 'model': 'Dummy Test Device',
+                'drivers': {
+                    'mint-dev-pkg': {'free': False, 'from_distro': True, 'recommended': True},
+                    'mint-dev-pkg-debconf': {'free': False, 'from_distro': True, 'recommended': False},
+                    'linux-generic': {'free': True, 'builtin': True, 'from_distro': True, 'recommended': False}}
+                }
+            self.devices['dummy'] = dummy_device
         self.show_drivers()
 
     @idle
@@ -592,9 +652,9 @@ class Application():
                 option_group = None
                 # define the order of introspection
                 for section in ('recommended', 'alternative', 'manually_installed', 'no_driver'):
-                    for driver in sorted(drivers[section], key=lambda x: self.sort_string(drivers[section], x)):
-                        if str(driver).startswith("nvidia-driver") and str(driver).endswith("-server"):
-                            print("Ignoring server NVIDIA driver: ", driver)
+                    for driver in sorted(drivers[section], key=lambda x: self.sort_string(drivers[section], x), reverse=True):
+                        if str(driver).startswith("nvidia-driver") and str(driver).endswith(("-server", "-open")):
+                            print("Ignoring server or open NVIDIA driver: ", driver)
                             continue
                         radio_button = Gtk.RadioButton.new(None)
                         label = Gtk.Label()
@@ -619,20 +679,19 @@ class Application():
                             radio_button.set_sensitive(False)
 
                 self.box_driver_detail.pack_start(device_box, False, False, 6)
-                self.stack.set_visible_child_name("drivers_page")
 
-        if not drivers_found:
-            self.stack.set_visible_child_name("no_drivers_page")
+        if drivers_found:
+            self.show_page("drivers_page")
+        else:
+            self.show_page("no_drivers_page")
             print("Your computer does not need any additional drivers")
 
-        self.spinner.stop()
-
         self.ui_building = False
         self.box_driver_detail.show_all()
         self.set_driver_action_status()
 
     def sort_string(self, drivers, x):
-        value = drivers[x]['description']
+        value = x
         try:
             value = "%s %s" % (not drivers[x]['free'], value)
         except:
@@ -640,7 +699,7 @@ class Application():
         return value
 
     def update_label_and_icons_from_status(self):
-        '''Update the current label and icon, computing the new device status'''
+        """Update the current label and icon, computing the new device status"""
 
         for device in self.devices:
             if device in self.dynamic_device_status.keys():
